#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>

#include "pkcs11_utils.h"

CK_RV compute_sha256(CK_FUNCTION_LIST_PTR function_list,
                     CK_SESSION_HANDLE session,
                     CK_BYTE_PTR data, CK_ULONG data_len,
                     CK_BYTE_PTR* digest_out, CK_ULONG* digest_len_out) {
    CK_RV rv;
    CK_MECHANISM mechanism = { CKM_SHA256, NULL_PTR, 0 };

    // 1. Initialize Digest
    rv = function_list->C_DigestInit(session, &mechanism);
    if (rv != CKR_OK) {
        fprintf(stderr, "C_DigestInit failed: 0x%lx\n", rv);
        return rv;
    }

    // 2. Provide data to digest
    rv = function_list->C_DigestUpdate(session, data, data_len);
    if (rv != CKR_OK) {
        fprintf(stderr, "C_DigestUpdate failed: 0x%lx\n", rv);
        return rv;
    }

    // 3. Get size of digest
    CK_ULONG digest_len = 0;
    rv = function_list->C_DigestFinal(session, NULL_PTR, &digest_len);
    if (rv != CKR_OK) {
        fprintf(stderr, "C_DigestFinal (size query) failed: 0x%lx\n", rv);
        return rv;
    }

    CK_BYTE_PTR digest = (CK_BYTE_PTR)malloc(digest_len);
    if (!digest) {
        fprintf(stderr, "Memory allocation for digest failed\n");
        return CKR_HOST_MEMORY;
    }

    // 4. Get actual digest
    rv = function_list->C_DigestFinal(session, digest, &digest_len);
    if (rv != CKR_OK) {
        fprintf(stderr, "C_DigestFinal failed: 0x%lx\n", rv);
        free(digest);
        return rv;
    }

    *digest_out = digest;
    *digest_len_out = digest_len;

    return CKR_OK;
}
     


int main(int argc, char** argv) {
    void *pkcs11_library = NULL;

    CK_FUNCTION_LIST_PTR function_list = NULL;

    CK_RV rv;
    CK_SLOT_ID slots[10];
    CK_ULONG slot_count;
    CK_SESSION_HANDLE session, soft_session;
    char *pin = "12345678";  // HSM PIN
    char *module_path = "./libikvp11.so";  // HSM module path

    pkcs11_library = load_pkcs11_library(module_path);
    if (!pkcs11_library) {
        return 1;
    }

    rv = get_function_list(pkcs11_library, &function_list);
    if (rv != CKR_OK) {
        dlclose(pkcs11_library);
        return 1;
    }


    rv = initialize_pkcs11(function_list);
    if (rv != CKR_OK) {
        dlclose(pkcs11_library);
        return 1;
    }


    rv = get_slots(function_list, slots, &slot_count);
    if (rv != CKR_OK) {
        finalize_pkcs11(function_list);
        dlclose(pkcs11_library);
        return 1;
    }


    if ( slot_count > 0) {

            rv = open_session(function_list, slots[0], &session);

            if (rv == CKR_OK) {

                CK_BYTE sample_data[] = "Testing data for sha256";
                CK_BYTE_PTR digest = NULL;
                CK_ULONG digest_len = 0;

                rv = compute_sha256(function_list, session,
                                    sample_data, sizeof(sample_data) - 1,
                                    &digest, &digest_len);

                if (rv == CKR_OK) {
                    printf("SHA-256 Digest (%lu bytes):\n", digest_len);
                    print_key_hex(digest, digest_len);
                    free(digest);
                }

            }
        
            close_session(function_list, session); 
        
    }

    finalize_pkcs11(function_list);

    dlclose(pkcs11_library);

    return 0;
}
