#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>

#include "pkcs11_utils.h"

#define RSA_KEY_SIZE 3072
#define AES_CKM  CKM_AES_KEY_WRAP_PAD
#define RSA_CKM CKM_RSA_PKCS_OAEP


const CK_RSA_PKCS_OAEP_PARAMS GLOBAL_OAEP_PARAMS = {  CKM_SHA_1, // Hash algorithm  
                                                      CKG_MGF1_SHA1, // Mask generation function  
                                                      CKZ_DATA_SPECIFIED, // Source type  
                                                      NULL,  
                                                      0 };  
#define INIT_RSA_OAEP_MECHANISM(mech_ptr) do { \
    (mech_ptr)->mechanism = RSA_CKM; \
    (mech_ptr)->pParameter = (void*)&GLOBAL_OAEP_PARAMS; \
    (mech_ptr)->ulParameterLen = sizeof(GLOBAL_OAEP_PARAMS); \
} while(0)

// Function declarations

CK_RV generate_extractable_key(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                              const char* key_label, CK_OBJECT_HANDLE_PTR key_handle);

CK_RV test_key_extraction(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                          const char* key_label);

CK_RV generate_and_verify_extractable_key(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                                          const char* key_label, CK_OBJECT_HANDLE_PTR key_handle);

CK_RV handle_aes_key_wrap(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                             const char* enc_key_label, const char* target_key_label,
                             CK_BYTE_PTR* wrapped_key_out, CK_ULONG* wrapped_key_out_len);

CK_RV extract_rsa_public_key(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                            CK_OBJECT_HANDLE public_key_handle,
                            CK_BYTE_PTR* modulus_out, CK_ULONG* modulus_len,
                            CK_BYTE_PTR* exponent_out, CK_ULONG* exponent_len);

CK_RV import_rsa_public_key_from_components(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                                           const char* key_label,
                                           CK_BYTE_PTR modulus, CK_ULONG modulus_len,
                                           CK_BYTE_PTR exponent, CK_ULONG exponent_len,
                                           CK_OBJECT_HANDLE_PTR key_handle);

CK_RV handle_rsa_key_wrapping_with_components(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                                               CK_BYTE_PTR modulus, CK_ULONG modulus_len,
                                               CK_BYTE_PTR exponent, CK_ULONG exponent_len,
                                               const char* target_key_label,
                                               CK_BYTE_PTR* wrapped_key_out, CK_ULONG* wrapped_key_out_len);

CK_RV generate_rsa_key_pair(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                           const char* public_key_label, const char* private_key_label,
                           CK_OBJECT_HANDLE_PTR public_key_handle, CK_OBJECT_HANDLE_PTR private_key_handle);


// ------- EXTRACTABLE KEY GENERATION -------

CK_RV generate_extractable_key(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                              const char* key_label, CK_OBJECT_HANDLE_PTR key_handle) {
    CK_OBJECT_CLASS key_class = CKO_SECRET_KEY;
    CK_KEY_TYPE key_type = CKK_AES;
    CK_BBOOL true_value = CK_TRUE;
    CK_BBOOL false_value = CK_FALSE;
    CK_ULONG key_size = 32; // 256 bits AES key


    CK_ATTRIBUTE template[] = {
        {CKA_CLASS, &key_class, sizeof(key_class)},
        {CKA_KEY_TYPE, &key_type, sizeof(key_type)},
        {CKA_LABEL, (void*)key_label, strlen(key_label)},
        {CKA_ID, (void*)key_label, strlen(key_label)},
        {CKA_TOKEN, &true_value, sizeof(true_value)},
        {CKA_PRIVATE, &false_value, sizeof(false_value)},     // Set to false to allow extraction
        {CKA_SENSITIVE, &false_value, sizeof(false_value)},   // Set to false to allow extraction
        {CKA_ENCRYPT, &true_value, sizeof(true_value)},
        {CKA_DECRYPT, &true_value, sizeof(true_value)},
        {CKA_EXTRACTABLE, &true_value, sizeof(true_value)},
        {CKA_WRAP, &true_value, sizeof(true_value)},
        {CKA_UNWRAP, &true_value, sizeof(true_value)},
        {CKA_DERIVE, &false_value, sizeof(false_value)},
        {CKA_VALUE_LEN, &key_size, sizeof(key_size)}
    };

    // Mechanism for AES key generation
    CK_MECHANISM mechanism = {
        CKM_AES_KEY_GEN, NULL, 0
    };

    CK_RV rv = function_list->C_GenerateKey(
        session,
        &mechanism,
        template,
        sizeof(template) / sizeof(CK_ATTRIBUTE),
        key_handle
    );

    if (rv != CKR_OK) {
        fprintf(stderr, "Error generating extractable key: 0x%lx\n", rv);
    } 
    return rv;
}

CK_RV test_key_extraction(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                          const char* key_label) {
    CK_RV rv;
    CK_OBJECT_HANDLE key_handle;
    CK_OBJECT_CLASS key_class = CKO_SECRET_KEY;
    CK_KEY_TYPE key_type = CKK_AES;
    CK_BBOOL tkn = CK_TRUE;
    // Find the key by label
    rv = find_key_by_label(function_list, session, key_label, key_class, &key_handle,key_type ,tkn);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error finding key '%s': 0x%lx\n", key_label, rv);
        return rv;
    }

    // Try to extract the key value
    unsigned char value_buffer[4096];
    CK_ATTRIBUTE get_value_template[] = {
        {CKA_VALUE, value_buffer, sizeof(value_buffer)}
    };

    rv = function_list->C_GetAttributeValue(session, key_handle, get_value_template, 1);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error extracting key value: 0x%lx\n", rv);
        return rv;
    }

    CK_ULONG key_length = get_value_template[0].ulValueLen;
    printf("Successfully extracted key value with label: %s, length: %lu bytes\n", key_label ,key_length);

    print_key_hex(value_buffer, key_length);


    return CKR_OK;
}

CK_RV generate_and_verify_extractable_key(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                                          const char* key_label, CK_OBJECT_HANDLE_PTR key_handle) {
    CK_RV rv;
    CK_OBJECT_CLASS key_class = CKO_SECRET_KEY;
    CK_KEY_TYPE key_type = CKK_AES;
    CK_BBOOL tkn = CK_TRUE;
    // First check if the key already exists
    if (is_label_in_use(function_list, session, key_label, key_class,key_type,tkn)) {
        printf("Key '%s' already exists. Attempting to verify it's extractable...\n", key_label);

        // Find the existing key
        rv = find_key_by_label(function_list, session, key_label, key_class, key_handle,key_type,tkn);
        if (rv != CKR_OK) {
            fprintf(stderr, "Error finding existing key: 0x%lx\n", rv);
            return rv;
        }

        // Test if it's extractable
        rv = test_key_extraction(function_list, session, key_label);
        if (rv == CKR_OK) {
            return CKR_OK;
        }

        printf("Existing key is not extractable. .\n");
        return rv;
    }
    printf("Generate the extractable key.\n");

    // Generate the extractable key
    rv = generate_extractable_key(function_list, session, key_label, key_handle);
    rv = test_key_extraction(function_list, session, key_label);
    if (rv != CKR_OK) {
        return rv;
    }	

    return rv;
}

CK_RV handle_aes_key_wrap(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                              const char* wrap_key_label, const char* target_key_label,
                              CK_BYTE_PTR* wrapped_key_out, CK_ULONG* wrapped_key_out_len) {
    CK_RV rv;
    CK_OBJECT_HANDLE enc_key_handle;
    CK_OBJECT_HANDLE target_key_handle;
    CK_OBJECT_CLASS secret_key_class = CKO_SECRET_KEY;
    CK_MECHANISM mechanism = {AES_CKM, NULL, 0};

    // 共
    rv = generate_and_verify_extractable_key(function_list, session, wrap_key_label, &enc_key_handle);
    if (rv != CKR_OK) {
        fprintf(stderr, "Failed to get 共 key: 0x%lx\n", rv);
        return rv;
    }

    // MK
    rv = generate_and_verify_extractable_key(function_list, session, target_key_label, &target_key_handle);
    if (rv != CKR_OK) {
        fprintf(stderr, "Failed to get MK key: 0x%lx\n", rv);
        return rv;
    }
 

    // Get the buffer size needed for the wrapped key
    CK_ULONG wrapped_key_len;
    rv = function_list->C_WrapKey(
        session,
        &mechanism,
        enc_key_handle,
        target_key_handle,
        NULL,
        &wrapped_key_len
    );

    if (rv != CKR_OK) {
        fprintf(stderr, "Error getting wrapped key size: 0x%lx\n", rv);
        return rv;
    }


    // Allocate buffer for the wrapped key
    CK_BYTE_PTR wrapped_key = (CK_BYTE_PTR)malloc(wrapped_key_len);
    if (wrapped_key == NULL) {
        fprintf(stderr, "Failed to allocate memory for wrapped key\n");
        return CKR_HOST_MEMORY;
    }

    // Perform the actual key wrapping
    rv = function_list->C_WrapKey(
        session,
        &mechanism,
        enc_key_handle,
        target_key_handle,
        wrapped_key,
        &wrapped_key_len
    );

    if (rv != CKR_OK) {
        fprintf(stderr, "Error in key wrapping: 0x%lx\n", rv);
        free(wrapped_key);
        return rv;
    }

    printf("Key wrapping completed successfully, length: %lu bytes\n", wrapped_key_len);

    // Copy the wrapped key to the output parameter
    *wrapped_key_out = (CK_BYTE_PTR)malloc(wrapped_key_len);
    if (*wrapped_key_out == NULL) {
        free(wrapped_key);
        return CKR_HOST_MEMORY;
    }

    memcpy(*wrapped_key_out, wrapped_key, wrapped_key_len);
    *wrapped_key_out_len = wrapped_key_len;

    free(wrapped_key);
    return CKR_OK;
}

CK_RV handle_rsa_key_wrapping_with_components(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                                               CK_BYTE_PTR modulus, CK_ULONG modulus_len,
                                               CK_BYTE_PTR exponent, CK_ULONG exponent_len,
                                               const char* target_key_label,
                                               CK_BYTE_PTR* wrapped_key_out, CK_ULONG* wrapped_key_out_len) {
    CK_RV rv;
    CK_OBJECT_HANDLE rsa_key_handle;
    CK_OBJECT_HANDLE target_key_handle;
    CK_OBJECT_CLASS secret_key_class = CKO_SECRET_KEY;
    const char* rsa_pubkey_label = "temp_rsa_key";
    CK_MECHANISM mechanism; 
    INIT_RSA_OAEP_MECHANISM(&mechanism);

    rv = import_rsa_public_key_from_components(
        function_list, session, rsa_pubkey_label,
        modulus, modulus_len, exponent, exponent_len,
        &rsa_key_handle
    );

    if (rv != CKR_OK) {
        fprintf(stderr, "Failed to import RSA key from components: 0x%lx\n", rv);
        return rv;
    }
    printf("Success to import RSA key from components \n");


    // Generate the target key to wrap, here the target key is kyouken key
    rv = generate_and_verify_extractable_key(function_list, session, target_key_label, &target_key_handle);
    if (rv != CKR_OK) {
        fprintf(stderr, "Failed to get kyouken key: 0x%lx\n", rv);
        return rv;
    }

    // Get the buffer size needed for the wrapped key
    CK_ULONG wrapped_key_len;
    rv = function_list->C_WrapKey(
        session,
        &mechanism,
        rsa_key_handle,
        target_key_handle,
        NULL,
        &wrapped_key_len
    );

    if (rv != CKR_OK) {
        fprintf(stderr, "Error getting RSA wrapped key size: 0x%lx\n", rv);
        return rv;
    }


    // Allocate buffer for the wrapped key
    CK_BYTE_PTR wrapped_key = (CK_BYTE_PTR)malloc(wrapped_key_len);
    if (wrapped_key == NULL) {
        fprintf(stderr, "Failed to allocate memory for wrapped key\n");
        return CKR_HOST_MEMORY;
    }

    // Perform the actual key wrapping
    rv = function_list->C_WrapKey(
        session,
        &mechanism,
        rsa_key_handle,
        target_key_handle,
        wrapped_key,
        &wrapped_key_len
    );

    if (rv != CKR_OK) {
        fprintf(stderr, "Error in RSA key wrapping: 0x%lx\n", rv);
        free(wrapped_key);
        return rv;
    }

    printf("RSA key wrapping completed successfully, length: %lu bytes\n", wrapped_key_len);

    // Copy the wrapped key to the output parameter
    *wrapped_key_out = (CK_BYTE_PTR)malloc(wrapped_key_len);
    if (*wrapped_key_out == NULL) {
        free(wrapped_key);
        return rv;
    }

    memcpy(*wrapped_key_out, wrapped_key, wrapped_key_len);
    *wrapped_key_out_len = wrapped_key_len;

    // Clean up the temporary RSA key if need
    // function_list->C_DestroyObject(session, rsa_key_handle);
    free(wrapped_key);

    return CKR_OK;
}

CK_RV unwrap_wrapped_keys(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                           CK_OBJECT_HANDLE rsa_private_key,
                           CK_BYTE_PTR wrapped_kyou_key, CK_ULONG wrapped_kyou_key_len,
                           CK_BYTE_PTR wrapped_master_key, CK_ULONG wrapped_master_key_len) {
    CK_RV rv;
    CK_OBJECT_HANDLE unwrapped_kyou_key_handle;
    CK_OBJECT_HANDLE unwrapped_master_key_handle;
    CK_OBJECT_CLASS secret_key_class = CKO_SECRET_KEY;
    CK_KEY_TYPE aes_key_type = CKK_AES;
    CK_BBOOL true_value = CK_TRUE;
    CK_BBOOL false_value = CK_FALSE;
    
    CK_MECHANISM rsa_mechanism; 
    INIT_RSA_OAEP_MECHANISM(&rsa_mechanism);

    CK_MECHANISM aes_mechanism = {AES_CKM, NULL, 0};
    char* unwrapped_key = "unwrapped_kyou_key";
    char* unwrapped_MK = "unwrapped_master_key";



    // 1. Unwrap the kyou key with RSA private key

    CK_ATTRIBUTE kyou_template[] = {
        {CKA_CLASS, &secret_key_class, sizeof(secret_key_class)},
        {CKA_KEY_TYPE, &aes_key_type, sizeof(aes_key_type)},
        {CKA_TOKEN, &true_value, sizeof(true_value)},
        {CKA_DECRYPT, &true_value, sizeof(true_value)},
        {CKA_UNWRAP, &true_value, sizeof(true_value)},
        {CKA_EXTRACTABLE, &true_value, sizeof(true_value)},
        {CKA_LABEL, unwrapped_key, strlen(unwrapped_key)}
    };

    rv = function_list->C_UnwrapKey(
        session,
        &rsa_mechanism,
        rsa_private_key,
        wrapped_kyou_key,
        wrapped_kyou_key_len,
        kyou_template,
        sizeof(kyou_template) / sizeof(CK_ATTRIBUTE),
        &unwrapped_kyou_key_handle
    );

    if (rv != CKR_OK) {
        fprintf(stderr, "Error unwrapping kyou key: 0x%lx\n", rv);
        return rv;
    }



    // 2. Extract and print the unwrapped kyou key value
    CK_ATTRIBUTE get_kyou_value[] = {
        {CKA_VALUE, NULL, 0}
    };

    rv = function_list->C_GetAttributeValue(session, unwrapped_kyou_key_handle,
                                           get_kyou_value, 1);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error getting kyou key value size: 0x%lx\n", rv);
        return rv;
    }

    get_kyou_value[0].pValue = malloc(get_kyou_value[0].ulValueLen);
    if (get_kyou_value[0].pValue == NULL) {
        fprintf(stderr, "Memory allocation failed for kyou key value\n");
        return CKR_HOST_MEMORY;
    }

    rv = function_list->C_GetAttributeValue(session, unwrapped_kyou_key_handle,
                                           get_kyou_value, 1);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error getting kyou key value: 0x%lx\n", rv);
        free(get_kyou_value[0].pValue);
        return rv;
    }
 
    printf("\nUnwrapped kyou key value (%lu bytes):\n", get_kyou_value[0].ulValueLen);
    print_key_hex(get_kyou_value[0].pValue, get_kyou_value[0].ulValueLen);
    free(get_kyou_value[0].pValue);
    // 3. Unwrap the master key using the kyou key
  

    CK_ATTRIBUTE master_key_template[] = {
        {CKA_CLASS, &secret_key_class, sizeof(secret_key_class)},
        {CKA_KEY_TYPE, &aes_key_type, sizeof(aes_key_type)},
        {CKA_TOKEN, &true_value, sizeof(true_value)},
        {CKA_DECRYPT, &true_value, sizeof(true_value)},
        {CKA_UNWRAP, &true_value, sizeof(true_value)},
        {CKA_EXTRACTABLE, &true_value, sizeof(true_value)},
        {CKA_LABEL, unwrapped_MK, strlen(unwrapped_MK)}
    };

    rv = function_list->C_UnwrapKey(
        session,
        &aes_mechanism,
        unwrapped_kyou_key_handle,
        wrapped_master_key,
        wrapped_master_key_len,
        master_key_template,
        sizeof(master_key_template) / sizeof(CK_ATTRIBUTE),
        &unwrapped_master_key_handle
    );

    if (rv != CKR_OK) {
        fprintf(stderr, "Error unwrapping master key: 0x%lx\n", rv);
        return rv;
    }


    // 4. Extract and print the unwrapped master key value
    CK_ATTRIBUTE get_master_value[] = {
        {CKA_VALUE, NULL, 0}
    };

    rv = function_list->C_GetAttributeValue(session, unwrapped_master_key_handle,
                                           get_master_value, 1);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error getting master key value size: 0x%lx\n", rv);
        return rv;
    }

    get_master_value[0].pValue = malloc(get_master_value[0].ulValueLen);
    if (get_master_value[0].pValue == NULL) {
        fprintf(stderr, "Memory allocation failed for master key value\n");
        return CKR_HOST_MEMORY;
    }

    rv = function_list->C_GetAttributeValue(session, unwrapped_master_key_handle,
                                           get_master_value, 1);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error getting master key value: 0x%lx\n", rv);
        free(get_master_value[0].pValue);
        return rv;
    }

    printf("\nUnwrapped master key value (%lu bytes):\n", get_master_value[0].ulValueLen);
    print_key_hex(get_master_value[0].pValue, get_master_value[0].ulValueLen);
    free(get_master_value[0].pValue);

    return CKR_OK;
}

CK_RV generate_rsa_key_pair(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                           const char* public_key_label, const char* private_key_label,
                           CK_OBJECT_HANDLE_PTR public_key_handle, CK_OBJECT_HANDLE_PTR private_key_handle) {
    CK_MECHANISM mechanism = {CKM_RSA_PKCS_KEY_PAIR_GEN, NULL, 0};
    CK_ULONG modulus_bits = RSA_KEY_SIZE;
    CK_BYTE public_exponent[] = {0x01, 0x00, 0x01}; // 65537
    CK_BBOOL true_value = CK_TRUE;
    CK_BBOOL false_value = CK_FALSE;
    CK_OBJECT_CLASS public_key_class = CKO_PUBLIC_KEY;
    CK_OBJECT_CLASS private_key_class = CKO_PRIVATE_KEY;
    CK_KEY_TYPE rsa_key_type = CKK_RSA;

    CK_ATTRIBUTE public_key_template[] = {
        {CKA_LABEL, (void*)public_key_label, strlen(public_key_label)},
        {CKA_ID, (void*)public_key_label, strlen(public_key_label)},
        {CKA_CLASS, &public_key_class, sizeof(public_key_class)},
        {CKA_KEY_TYPE, &rsa_key_type, sizeof(rsa_key_type)},
        {CKA_TOKEN, &true_value, sizeof(true_value)},
        {CKA_ENCRYPT, &true_value, sizeof(true_value)},
        {CKA_VERIFY, &true_value, sizeof(true_value)},
        {CKA_WRAP, &true_value, sizeof(true_value)},
        {CKA_MODULUS_BITS, &modulus_bits, sizeof(modulus_bits)},
        {CKA_PUBLIC_EXPONENT, public_exponent, sizeof(public_exponent)},
    };

    CK_ATTRIBUTE private_key_template[] = {
        {CKA_CLASS, &private_key_class, sizeof(private_key_class)},
        {CKA_KEY_TYPE, &rsa_key_type, sizeof(rsa_key_type)},
        {CKA_TOKEN, &true_value, sizeof(true_value)},
        {CKA_PRIVATE, &true_value, sizeof(true_value)},
        {CKA_SENSITIVE, &true_value, sizeof(true_value)},
        {CKA_DECRYPT, &true_value, sizeof(true_value)},
        {CKA_UNWRAP, &true_value, sizeof(true_value)},
        {CKA_EXTRACTABLE, &false_value, sizeof(false_value)},
        {CKA_LABEL, (void*)private_key_label, strlen(private_key_label)},
        {CKA_ID, (void*)private_key_label, strlen(private_key_label)},
    };

    printf("Generating RSA key pair (%lu bits)...\n", modulus_bits);
    CK_RV rv = function_list->C_GenerateKeyPair(
        session,
        &mechanism,
        public_key_template, sizeof(public_key_template) / sizeof(CK_ATTRIBUTE),
        private_key_template, sizeof(private_key_template) / sizeof(CK_ATTRIBUTE),
        public_key_handle,
        private_key_handle
    );

    if (rv != CKR_OK) {
        fprintf(stderr, "Error generating RSA key pair: 0x%lx\n", rv);
    } else {
        printf("RSA key pair generated successfully.\n");
    }

    return rv;
}

CK_RV extract_rsa_public_key(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                            CK_OBJECT_HANDLE public_key_handle,
                            CK_BYTE_PTR* modulus_out, CK_ULONG* modulus_len,
                            CK_BYTE_PTR* exponent_out, CK_ULONG* exponent_len) {
    CK_RV rv;

    // First get the size of modulus
    CK_ATTRIBUTE modulus_template[] = {
        {CKA_MODULUS, NULL, 0}
    };

    rv = function_list->C_GetAttributeValue(session, public_key_handle, modulus_template, 1);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error getting modulus size: 0x%lx\n", rv);
        return rv;
    }

    // Allocate memory for modulus
    *modulus_len = modulus_template[0].ulValueLen;
    *modulus_out = (CK_BYTE_PTR)malloc(*modulus_len);
    if (*modulus_out == NULL) {
        fprintf(stderr, "Memory allocation failed for modulus\n");
        return CKR_HOST_MEMORY;
    }

    modulus_template[0].pValue = *modulus_out;

    // Now get the exponent size
    CK_ATTRIBUTE exponent_template[] = {
        {CKA_PUBLIC_EXPONENT, NULL, 0}
    };

    rv = function_list->C_GetAttributeValue(session, public_key_handle, exponent_template, 1);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error getting exponent size: 0x%lx\n", rv);
        free(*modulus_out);
        *modulus_out = NULL;
        return rv;
    }

    // Allocate memory for exponent
    *exponent_len = exponent_template[0].ulValueLen;
    *exponent_out = (CK_BYTE_PTR)malloc(*exponent_len);
    if (*exponent_out == NULL) {
        fprintf(stderr, "Memory allocation failed for exponent\n");
        free(*modulus_out);
        *modulus_out = NULL;
        return CKR_HOST_MEMORY;
    }

    exponent_template[0].pValue = *exponent_out;

    // Get the actual values
    rv = function_list->C_GetAttributeValue(session, public_key_handle, modulus_template, 1);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error getting modulus value: 0x%lx\n", rv);
        free(*modulus_out);
        free(*exponent_out);
        *modulus_out = NULL;
        *exponent_out = NULL;
        return rv;
    }

    rv = function_list->C_GetAttributeValue(session, public_key_handle, exponent_template, 1);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error getting exponent value: 0x%lx\n", rv);
        free(*modulus_out);
        free(*exponent_out);
        *modulus_out = NULL;
        *exponent_out = NULL;
        return rv;
    }

    printf("RSA public key extracted successfully:\n");

    return CKR_OK;
}


CK_RV import_rsa_public_key_from_components(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                                           const char* key_label,
                                           CK_BYTE_PTR modulus, CK_ULONG modulus_len,
                                           CK_BYTE_PTR exponent, CK_ULONG exponent_len,
                                           CK_OBJECT_HANDLE_PTR key_handle) {
    CK_OBJECT_CLASS key_class = CKO_PUBLIC_KEY;
    CK_KEY_TYPE key_type = CKK_RSA;
    CK_BBOOL true_value = CK_TRUE;
    CK_BBOOL false_value = CK_FALSE;

    CK_ATTRIBUTE template[] = {
        {CKA_CLASS, &key_class, sizeof(key_class)},
        {CKA_KEY_TYPE, &key_type, sizeof(key_type)},
        {CKA_LABEL, (void*)key_label, strlen(key_label)},
        {CKA_ID, (void*)key_label, strlen(key_label)},
        {CKA_TOKEN, &true_value, sizeof(true_value)},
        {CKA_ENCRYPT, &true_value, sizeof(true_value)},
        {CKA_VERIFY, &true_value, sizeof(true_value)},
        {CKA_MODULUS, modulus, modulus_len},
        {CKA_PUBLIC_EXPONENT, exponent, exponent_len}
    };

    CK_RV rv = function_list->C_CreateObject(session, template,
                                         sizeof(template) / sizeof(CK_ATTRIBUTE),
                                         key_handle);

    if (rv != CKR_OK) {
        fprintf(stderr, "Error creating RSA public key: 0x%lx\n", rv);
    } else {
        printf("RSA public key imported successfully with handle: 0x%lx\n", *key_handle);
    }

    return rv;
}



int main(int argc, char** argv) {
    void *pkcs11_library = NULL;
    void *soft_pkcs11_library = NULL;
    CK_FUNCTION_LIST_PTR function_list = NULL;
    CK_FUNCTION_LIST_PTR soft_function_list= NULL;
    CK_RV rv;
    CK_SLOT_ID slots[10], soft_slots[10];
    CK_ULONG slot_count, soft_slot_count = 10;
    CK_SESSION_HANDLE session, soft_session;
    char *soft_pin = "12345678";  // softhsm PIN
    char *pin = "12345678";  // HSM PIN
    char *soft_module_path = "/usr/lib/softhsm/libsofthsm2.so";  // softhsm module path
    char *module_path = "./libikvp11.so";  // HSM module path
    int i;

    pkcs11_library = load_pkcs11_library(module_path);
    if (!pkcs11_library) {
        return 1;
    }

    // softhsm
    soft_pkcs11_library = load_pkcs11_library(soft_module_path);
    if (!soft_pkcs11_library) {
        return 1;
    }

    rv = get_function_list(pkcs11_library, &function_list);
    if (rv != CKR_OK) {
        dlclose(pkcs11_library);
        return 1;
    }

    // softhsm
    rv = get_function_list(soft_pkcs11_library, &soft_function_list);
    if (rv != CKR_OK) {
        dlclose(soft_pkcs11_library);
        return 1;
    }

    rv = initialize_pkcs11(function_list);
    if (rv != CKR_OK) {
        dlclose(pkcs11_library);
        return 1;
    }

    // softhsm
    rv = initialize_pkcs11(soft_function_list);
    if (rv != CKR_OK) {
        dlclose(soft_pkcs11_library);
        return 1;
    }

    rv = get_slots(function_list, slots, &slot_count);
    if (rv != CKR_OK) {
        finalize_pkcs11(function_list);
        dlclose(pkcs11_library);
        return 1;
    }

    // softhsm
    rv = get_slots(soft_function_list, soft_slots, &soft_slot_count);
    if (rv != CKR_OK) {
        finalize_pkcs11(soft_function_list);
        dlclose(soft_pkcs11_library);
        return 1;
    }

    if ( soft_slot_count > 0) {
        rv = open_session(function_list, slots[0], &session);
        // softhsm
        rv = open_session(soft_function_list, soft_slots[0], &soft_session);

        if (rv == CKR_OK) {
            rv = login(function_list, session, CKU_USER, pin);
            // softhsm
            rv = login(soft_function_list, soft_session, CKU_USER, soft_pin);
            if (rv == CKR_OK) {

    printf("\n ------------- softhsm pretend SERVER generate RSA key pair ----------- \n");
                // 1. SERVER Generate RSA key pair first
                CK_OBJECT_HANDLE rsa_public_key_handle, rsa_private_key_handle;

                rv = generate_rsa_key_pair(soft_function_list, soft_session,
                                          "srv_rsa_pub_key", "srv_rsa_priv_key",
                                          &rsa_public_key_handle, &rsa_private_key_handle);

                if (rv != CKR_OK) {
                    fprintf(stderr, "Failed to generate RSA key pair: 0x%lx\n", rv);
                    logout(soft_function_list, soft_session);
                    close_session(soft_function_list, soft_session);
                    finalize_pkcs11(soft_function_list);
                    dlclose(soft_pkcs11_library);
                    return 1;
                }

                // 2. Extract RSA public key components
                CK_BYTE_PTR modulus = NULL;
                CK_ULONG modulus_len = 0;
                CK_BYTE_PTR exponent = NULL;
                CK_ULONG exponent_len = 0;

                rv = extract_rsa_public_key(soft_function_list, soft_session,
                                          rsa_public_key_handle,
                                          &modulus, &modulus_len,
                                          &exponent, &exponent_len);

                if (rv != CKR_OK) {
                    fprintf(stderr, "Failed to extract RSA key components: 0x%lx\n", rv);
                    logout(soft_function_list, soft_session);
                    close_session(soft_function_list, soft_session);
                    finalize_pkcs11(soft_function_list);
                    dlclose(soft_pkcs11_library);
                    return 1;
                }
    printf("\n ------------- softhsm Pretend SERVER generate RSA key pair END----------- \n");

                CK_BYTE_PTR wrapped_master_key = NULL;
                CK_ULONG wrapped_master_key_len = 0;
                CK_BYTE_PTR wrapped_kyou_key = NULL;
                CK_ULONG wrapped_kyou_key_len = 0;

                // 3. HSM Generating kyouken_key and wrap MK
                printf("\n Wrap Mk key with kyouken...\n");
                rv = handle_aes_key_wrap(function_list, session,
                                           "kyouken_key", "MK_key",
                                           &wrapped_master_key, &wrapped_master_key_len);

                // 4. HSM Import RSA pub and wrap kyouken_key
                printf("\n Wrap kyouken with RSA pub...\n");
                rv = handle_rsa_key_wrapping_with_components(
                    function_list, session,
                    modulus, modulus_len, exponent, exponent_len,
                    "kyouken_key",
                    &wrapped_kyou_key, &wrapped_kyou_key_len
                );
                printf("\n ///////// print wrapped_master_key /////// \n");

                print_key_hex(wrapped_master_key, wrapped_master_key_len);

                printf("\n //////// print wrapped_kyou_key ////// \n");
                print_key_hex(wrapped_kyou_key, wrapped_kyou_key_len);


    printf("\n ------------- softhsm pretend SERVER unwrap MK ----------- \n");
                // 5. export wrapped_master_key wrapped_kyou_key to SERVER
                // 6. unwrap the wrapped keys to get MK and kyou key
                rv = unwrap_wrapped_keys(soft_function_list, soft_session,
                                          rsa_private_key_handle,
                                          wrapped_kyou_key, wrapped_kyou_key_len,
                                          wrapped_master_key, wrapped_master_key_len);

                if (wrapped_master_key) free(wrapped_master_key);
                if (wrapped_kyou_key) free(wrapped_kyou_key);
    printf("\n ------------- softhsm pretend SERVER unwrap MK END----------- \n");
                logout(function_list, session);
                logout(soft_function_list, soft_session);
            }
            close_session(function_list, session);
            close_session(soft_function_list, soft_session);
        }
    }

    finalize_pkcs11(function_list);
    finalize_pkcs11(soft_function_list);

    dlclose(pkcs11_library);
    dlclose(soft_pkcs11_library);

    return 0;
}
