#include "pkcs11_utils.h"



void print_key_hex(CK_BYTE_PTR key_data, CK_ULONG key_len) {
    for (CK_ULONG i = 0; i < key_len; i++) {
        printf("%02X", key_data[i]);
        if ((i+1) % 16 == 0) printf("\n");
        else if ((i+1) % 8 == 0) printf("  ");
        else printf(" ");
    }
    printf("\n");
}

void* load_pkcs11_library(const char* library_path) {
    void* pkcs11_library = dlopen(library_path, RTLD_LAZY);
    if (!pkcs11_library) {
        fprintf(stderr, "Error loading library: %s\n", dlerror());
    }
    return pkcs11_library;
}

// Function to get the PKCS#11 function list
CK_RV get_function_list(void* pkcs11_library, CK_FUNCTION_LIST_PTR* function_list) {
    CK_C_GetFunctionList pfnGetFunctionList = NULL;
    CK_RV rv;

    pfnGetFunctionList = (CK_C_GetFunctionList)dlsym(pkcs11_library, "C_GetFunctionList");
    if (!pfnGetFunctionList) {
        fprintf(stderr, "Error finding C_GetFunctionList: %s\n", dlerror());
        return CKR_FUNCTION_FAILED;
    }

    rv = pfnGetFunctionList((CK_FUNCTION_LIST_3_0 **)function_list);
    if (rv != CKR_OK || !*function_list) {
        fprintf(stderr, "Error getting function list: 0x%lx\n", rv);
    }
    return rv;
}

// Function to initialize the PKCS#11 library
CK_RV initialize_pkcs11(CK_FUNCTION_LIST_PTR function_list) {
    CK_RV rv = function_list->C_Initialize(NULL);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error initializing library: 0x%lx\n", rv);
    }
    return rv;
}

// Function to finalize the PKCS#11 library
CK_RV finalize_pkcs11(CK_FUNCTION_LIST_PTR function_list) {
    CK_RV rv = function_list->C_Finalize(NULL);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error finalizing library: 0x%lx\n", rv);
    }
    return rv;
}


CK_RV get_slots(CK_FUNCTION_LIST_PTR function_list, CK_SLOT_ID* slots, CK_ULONG* slot_count) {
    CK_RV rv = function_list->C_GetSlotList(CK_TRUE, slots, slot_count);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error getting slot list: 0x%lx\n", rv);
    } else {
        printf("Found %lu slots\n", *slot_count);
        //GettokenInfo
    CK_TOKEN_INFO tokeninfo;
    rv = function_list->C_GetTokenInfo(slots[0],&tokeninfo);
    if (rv != CKR_OK){
        printf("C_GetTokeninfo failed with error 0x%lx",rv);
        return rv;
    }
    printf("Token Label: %s\n",tokeninfo.label);
    }
    return rv;
}

// Function to open a session
CK_RV open_session(CK_FUNCTION_LIST_PTR function_list, CK_SLOT_ID slot_id, CK_SESSION_HANDLE* session) {
    CK_RV rv = function_list->C_OpenSession(slot_id, CKF_SERIAL_SESSION | CKF_RW_SESSION,
                                           NULL, NULL, session);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error opening session: 0x%lx\n", rv);
    } else {
        printf("Session opened successfully\n");
    }
    return rv;
}

// Function to close a session
CK_RV close_session(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session) {
    CK_RV rv = function_list->C_CloseSession(session);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error closing session: 0x%lx\n", rv);
    }
    return rv;
}

// Function to log in to a session
CK_RV login(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session, CK_USER_TYPE user_type, const char* pin) {
    CK_UTF8CHAR* pin_bytes = (CK_UTF8CHAR*)pin;
    CK_RV rv = function_list->C_Login(session, user_type, pin_bytes, strlen(pin));
    if (rv != CKR_OK) {
        fprintf(stderr, "Error logging in: 0x%lx\n", rv);
    } else {
        printf("Logged in successfully\n");
    }
    return rv;
}

// Function to log out from a session
CK_RV logout(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session) {
    CK_RV rv = function_list->C_Logout(session);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error logging out: 0x%lx\n", rv);
    }
    return rv;
}

CK_BBOOL is_label_in_use(CK_FUNCTION_LIST_PTR function_list,
                            CK_SESSION_HANDLE session, 
                            const char* key_label, 
                            CK_OBJECT_CLASS key_class,
                            CK_KEY_TYPE key_type, 
                            CK_BBOOL tkn) {

    CK_ATTRIBUTE template[] = {
        {CKA_CLASS, &key_class, sizeof(key_class)},
        {CKA_LABEL, (void*)key_label, strlen(key_label)},
        {CKA_ID, (void*)key_label, strlen(key_label)},
        {CKA_KEY_TYPE, &key_type, sizeof(CK_KEY_TYPE) },
        {CKA_TOKEN,  &tkn, sizeof(CK_BBOOL) },
    };

    CK_OBJECT_HANDLE object_handle;
    CK_ULONG found_count;
    CK_RV rv;


    rv = function_list->C_FindObjectsInit(session, template, sizeof(template) / sizeof(CK_ATTRIBUTE));
    if (rv != CKR_OK) {
        fprintf(stderr, "Error initializing object search: 0x%lx\n", rv);
        return CK_FALSE;
    }


    rv = function_list->C_FindObjects(session, &object_handle, 1, &found_count);
    if (rv != CKR_OK) {
        fprintf(stderr, "Error finding objects: 0x%lx\n", rv);
        function_list->C_FindObjectsFinal(session);
        return CK_FALSE;
    }

    function_list->C_FindObjectsFinal(session);
    printf(" finding found_count: 0x%ld\n", found_count);

    return (found_count > 0) ? CK_TRUE : CK_FALSE;
}


CK_RV find_key_by_label(CK_FUNCTION_LIST_PTR function_list, CK_SESSION_HANDLE session,
                        const char* key_label, CK_OBJECT_CLASS key_class, CK_OBJECT_HANDLE_PTR key_handle,
                        CK_KEY_TYPE key_type, 
                        CK_BBOOL tkn) 
{
    CK_ATTRIBUTE template[] = {
    {CKA_CLASS, &key_class, sizeof(key_class)},
    {CKA_LABEL, (void*)key_label, strlen(key_label)},
    { CKA_ID, (void*)key_label, strlen(key_label)},
    { CKA_TOKEN,  &tkn, sizeof(CK_BBOOL) },
    {CKA_KEY_TYPE, &key_type, sizeof(CK_KEY_TYPE) },


    };

    CK_ULONG found_count;
    CK_RV rv;

    rv = function_list->C_FindObjectsInit(session, template, sizeof(template) / sizeof(CK_ATTRIBUTE));
    if (rv != CKR_OK) {
    fprintf(stderr, "Error initializing object search: 0x%lx\n", rv);
    return rv;
    }

    rv = function_list->C_FindObjects(session, key_handle, 1, &found_count);
    if (rv != CKR_OK) {
    fprintf(stderr, "Error finding objects: 0x%lx\n", rv);
    function_list->C_FindObjectsFinal(session);
    return rv;
    }

    rv = function_list->C_FindObjectsFinal(session);
    if (rv != CKR_OK) {
    fprintf(stderr, "Error finalizing object search: 0x%lx\n", rv);
    return rv;
    }

    if (found_count == 0) {
    fprintf(stderr, "Key with label '%s' not found\n", key_label);
    return CKR_OBJECT_HANDLE_INVALID;
    }

    return CKR_OK;
}